# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module that pushes input to ffmpeg to transcode into various formats."""

import shlex

from streamer.bitrate_configuration import AudioCodec, VideoCodec
from streamer.input_configuration import Input, InputType, MediaType
from streamer.node_base import PolitelyWaitOnFinish
from streamer.output_stream import AudioOutputStream, OutputStream, TextOutputStream, VideoOutputStream
from streamer.pipeline_configuration import PipelineConfig, StreamingMode
from typing import List, Union, Optional

class TranscoderNode(PolitelyWaitOnFinish):

  def __init__(self,
               inputs: List[Input],
               pipeline_config: PipelineConfig,
               outputs: List[OutputStream],
               index: int,
               hermetic_ffmpeg: Optional[str]) -> None:
    super().__init__()
    self._inputs = inputs
    self._pipeline_config = pipeline_config
    self._outputs = outputs
    self._index = index
    # If a hermetic ffmpeg is passed, use it.
    self._ffmpeg = hermetic_ffmpeg or 'ffmpeg'

  def start(self) -> None:
    args = [
        self._ffmpeg,
        # Do not prompt for output files that already exist. Since we created
        # the named pipe in advance, it definitely already exists. A prompt
        # would block ffmpeg to wait for user input.
        '-y',
    ]

    if self._pipeline_config.quiet:
      args += [
          # Suppresses all messages except errors.
          # Without this, a status line will be printed by default showing
          # progress and transcoding speed.
          '-loglevel', 'error',
      ]

    if any([output.is_hardware_accelerated() for output in self._outputs]):
      if self._pipeline_config.hwaccel_api == 'vaapi':
        args += [
            # Hardware acceleration args.
            # TODO(#17): Support multiple VAAPI devices.
            '-vaapi_device', '/dev/dri/renderD128',
        ]

    for input in self._inputs:
      # Get any required input arguments for this input.
      # These are like hard-coded extra_input_args for certain input types.
      # This means users don't have to know much about FFmpeg options to handle
      # these common cases.
      args += input.get_input_args()

      # The config file may specify additional args needed for this input.
      # This allows, for example, an external-command-type input to generate
      # almost anything ffmpeg could ingest.  The extra args need to be parsed
      # from a string into an argument array.  Note that shlex.split on an empty
      # string will produce an empty array.
      args += shlex.split(input.extra_input_args)

      if input.input_type == InputType.LOOPED_FILE:
        # These are handled here instead of in get_input_args() because these
        # arguments are specific to ffmpeg and are not understood by ffprobe.
        args += [
            # Loop the input forever.
            '-stream_loop', '-1',
            # Read input in real time; don't go above 1x processing speed.
            '-re',
        ]

      if self._pipeline_config.streaming_mode == StreamingMode.LIVE:
        args += [
            # A larger queue to buffer input from the pipeline (default is 8).
            # This is in packets, but for raw images, that means frames.  A
            # 720p PPM frame is 2.7MB, and a 1080p PPM is 6.2MB.  The entire
            # queue, when full, must fit into memory.
            '-thread_queue_size', '200',
        ]

      if input.start_time:
        args += [
            # Encode from intended starting time of the input.
            '-ss', input.start_time,
        ]
      if input.end_time:
        args += [
            # Encode until intended ending time of the input.
            '-to', input.end_time,
        ]

      # The input name always comes after the applicable input arguments.
      args += [
          # The input itself.
          '-i', input.name,
      ]

    for i, input in enumerate(self._inputs):
      map_args = [
          # Map corresponding input stream to output file.
          # The format is "<INPUT FILE NUMBER>:<STREAM SPECIFIER>", so "i" here
          # is the input file number, and "input.get_stream_specifier()" builds
          # the stream specifier for this input.  The output stream for this
          # input is implied by where we are in the ffmpeg argument list.
          '-map', '{0}:{1}'.format(i, input.get_stream_specifier()),
      ]

      for output_stream in self._outputs:
        if output_stream.input != input:
          # Skip outputs that don't match this exact input object.
          continue
        if output_stream.skip_transcoding:
          # This input won't be transcoded.  This is common for VTT text input.
          continue

        # Map arguments must be repeated for each output file.
        args += map_args

        if input.media_type == MediaType.AUDIO:
          assert(isinstance(output_stream, AudioOutputStream))
          args += self._encode_audio(output_stream, input)
        elif input.media_type == MediaType.VIDEO:
          assert(isinstance(output_stream, VideoOutputStream))
          args += self._encode_video(output_stream, input)
        else:
          assert(isinstance(output_stream, TextOutputStream))
          args += self._encode_text(output_stream, input)

        args += [output_stream.ipc_pipe.write_end()]

    env = {}
    if self._pipeline_config.debug_logs:
      # Use this environment variable to turn on ffmpeg's logging.  This is
      # independent of the -loglevel switch above.
      ffmpeg_log_file = 'TranscoderNode-' + str(self._index) + '.log'
      env['FFREPORT'] = 'file={}:level=32'.format(ffmpeg_log_file)

    self._process = self._create_process(args, env)

  def _encode_audio(self, stream: AudioOutputStream, input: Input) -> List[str]:
    filters: List[str] = []
    args: List[str] = [
        # No video encoding for audio.
        '-vn',
        # TODO: This implied downmixing is not ideal.
        # Set the number of channels to the one specified in the config.
        '-ac', str(stream.layout.max_channels),
    ]

    if stream.layout.max_channels == 6:
      filters += [
        # Work around for https://github.com/shaka-project/shaka-packager/issues/598,
        # as seen on https://trac.ffmpeg.org/ticket/6974
        'channelmap=channel_layout=5.1',
      ]

    filters.extend(input.filters)

    hwaccel_api = self._pipeline_config.hwaccel_api
    args += [
        # Set codec and bitrate.
        '-c:a', stream.get_ffmpeg_codec_string(hwaccel_api),
        '-b:a', stream.get_bitrate(),
        # Output MP4 in the pipe, for all codecs.
        '-f', 'mp4',
        # This explicit fragment duration affects both audio and video, and
        # ensures that there are no single large MP4 boxes that Shaka Packager
        # can't consume from a pipe.
        # FFmpeg fragment duration is in microseconds.
        '-frag_duration', str(self._pipeline_config.segment_size * 1e6),
        # Opus in MP4 is considered "experimental".
        '-strict', 'experimental',
      ]

    if len(filters):
      args += [
          # Set audio filters.
          '-af', ','.join(filters),
      ]

    return args

  def _encode_video(self, stream: VideoOutputStream, input: Input) -> List[str]:
    filters: List[str] = []
    args: List[str] = []

    if input.is_interlaced:
      filters.append('pp=fd')
      args.extend(['-r', str(input.frame_rate)])
    
    if stream.resolution.max_frame_rate < input.frame_rate:
       args.extend(['-r', str(stream.resolution.max_frame_rate)])

    filters.extend(input.filters)

    hwaccel_api = self._pipeline_config.hwaccel_api

    # -2 in the scale filters means to choose a value to keep the original
    # aspect ratio.
    if stream.is_hardware_accelerated() and hwaccel_api == 'vaapi':
      # These filters are specific to Linux's vaapi.
      filters.append('format=nv12')
      filters.append('hwupload')
      filters.append('scale_vaapi=-2:{0}'.format(stream.resolution.max_height))
    else:
      filters.append('scale=-2:{0}'.format(stream.resolution.max_height))

    # To avoid weird rounding errors in Sample Aspect Ratio, set it explicitly
    # to 1:1.  Without this, you wind up with SAR set to weird values in DASH
    # that are very close to 1, such as 5120:5123.  In HLS, the behavior is
    # worse.  Some of the width values in the playlist wind up off by one,
    # which causes playback failures in ExoPlayer.
    # https://github.com/shaka-project/shaka-streamer/issues/36
    filters.append('setsar=1:1')

    if (stream.codec in {VideoCodec.H264, VideoCodec.HEVC} 
        and not stream.is_hardware_accelerated()):
      # These presets are specifically recognized by the software encoder.
      if self._pipeline_config.streaming_mode == StreamingMode.LIVE:
        args += [
            # Encodes with highest-speed presets for real-time live streaming.
            '-preset', 'ultrafast',
        ]
      else:
        args += [
            # Take your time for VOD streams.
            '-preset', 'slow',
            # Apply the loop filter for higher quality output.
            '-flags', '+loop',
        ]

    if stream.codec == VideoCodec.H264:  # Software or hardware
      # Use the "high" profile for HD and up, and "main" for everything else.
      # https://en.wikipedia.org/wiki/Advanced_Video_Coding#Profiles
      if stream.resolution.max_height >= 720:
        profile = 'high'
      else:
        profile = 'main'

      args += [
          # Set the H264 profile.  Without this, the default would be "main".
          # Note that this gets overridden to "baseline" in live streams by the
          # "-preset ultrafast" option, presumably because the baseline encoder
          # is faster.
          '-profile:v', profile,
      ]
      
    if stream.codec in {VideoCodec.H264, VideoCodec.HEVC}:
      args += [
          # The only format supported by QT/Apple.
          '-pix_fmt', 'yuv420p',
          # Require a closed GOP.  Some decoders don't support open GOPs.
          '-flags', '+cgop',
         
      ]

    elif stream.codec == VideoCodec.VP9:
      # TODO: Does -preset apply here?
      args += [
          # According to the wiki (https://trac.ffmpeg.org/wiki/Encode/VP9),
          # this allows threaded encoding in VP9, which makes better use of CPU
          # resources and speeds up encoding.  This is still not the default
          # setting as of libvpx v1.7.
          '-row-mt', '1',
          # speeds up encoding, balancing against quality
          '-speed', '2',
      ]
    elif stream.codec == VideoCodec.AV1:
      args += [
          # According to graphs at https://bit.ly/2BmIVt6, this AV1 setting
          # results in almost no reduction in quality (0.8%), but a significant
          # boost in speed (20x).
          '-cpu-used', '8',
          # According to the wiki (https://trac.ffmpeg.org/wiki/Encode/AV1),
          # this allows threaded encoding in AV1, which makes better use of CPU
          # resources and speeds up encoding.
          '-row-mt', '1',
      ]

    keyframe_interval = int(self._pipeline_config.segment_size *
                            input.frame_rate)

    args += [
        # No audio encoding for video.
        '-an',
        # Set codec and bitrate.
        '-c:v', stream.get_ffmpeg_codec_string(hwaccel_api),
        '-b:v', stream.get_bitrate(),
        # Output MP4 in the pipe, for all codecs.
        '-f', 'mp4',
        # This flag forces a video fragment at each keyframe.
        '-movflags', '+frag_keyframe',
        # This explicit fragment duration affects both audio and video, and
        # ensures that there are no single large MP4 boxes that Shaka Packager
        # can't consume from a pipe.
        # FFmpeg fragment duration is in microseconds.
        '-frag_duration', str(self._pipeline_config.segment_size * 1e6),
        # Set minimum and maximum GOP length.
        '-keyint_min', str(keyframe_interval), '-g', str(keyframe_interval),
        # Set video filters.
        '-vf', ','.join(filters),
    ]
    return args

  def _encode_text(self, stream: TextOutputStream, input: Input) -> List[str]:
    return [
        # Output WebVTT.
        '-f', 'webvtt',
    ]



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Top-level module API.

If you'd like to import Shaka Streamer as a Python module and build it into
your own application, this is the top-level API you can use for that.  You may
also want to look at the source code to the command-line front end script
`shaka-streamer`.
"""


import os
import re
import shutil
import subprocess
import sys
import tempfile

from typing import Any, Dict, List, Optional, Tuple, Union
from streamer import __version__
from streamer import autodetect
from streamer import min_versions
from streamer.bitrate_configuration import BitrateConfig, AudioChannelLayout, VideoResolution
from streamer.external_command_node import ExternalCommandNode
from streamer.input_configuration import InputConfig, InputType, MediaType, Input
from streamer.node_base import NodeBase, ProcessStatus
from streamer.output_stream import AudioOutputStream, OutputStream, TextOutputStream, VideoOutputStream
from streamer.packager_node import PackagerNode
from streamer.pipeline_configuration import ManifestFormat, PipelineConfig, StreamingMode
from streamer.transcoder_node import TranscoderNode
from streamer.periodconcat_node import PeriodConcatNode
from streamer.proxy_node import ProxyNode
import streamer.subprocessWindowsPatch  # side-effects only
from streamer.util import is_http_url, is_url
from streamer.pipe import Pipe


class ControllerNode(object):
  """Controls all other nodes and manages shared resources."""

  def __init__(self) -> None:
    global_temp_dir = tempfile.gettempdir()

    # The docs state that if any of prefix, suffix, or dir are specified, all
    # must be specified (and not None).  Create a temp dir of our own, inside
    # the global temp dir, and with a name that indicates who made it.
    self._temp_dir: str = tempfile.mkdtemp(
        dir=global_temp_dir, prefix='shaka-live-', suffix='')

    self._nodes: List[NodeBase] = []

  def __del__(self) -> None:
    # Clean up named pipes by removing the temp directory we placed them in.
    shutil.rmtree(self._temp_dir)

  def __enter__(self) -> 'ControllerNode':
    return self

  def __exit__(self, *unused_args) -> None:
    self.stop()

  def start(self, output_location: str,
            input_config_dict: Dict[str, Any],
            pipeline_config_dict: Dict[str, Any],
            bitrate_config_dict: Dict[Any, Any] = {},
            check_deps: bool = True,
            use_hermetic: bool = True) -> 'ControllerNode':
    """Create and start all other nodes.

    :raises: `RuntimeError` if the controller has already started.
    :raises: :class:`streamer.configuration.ConfigError` if the configuration is
             invalid.
    """

    if use_hermetic:
      try:
        import streamer_binaries # type: ignore
      except ImportError as ex:
        # If the package couldn't be imported.
        raise RuntimeError(
            'shaka-streamer-binaries was not found.\n'
            '  Install it with `pip install shaka-streamer-binaries`.\n'
            '  Alternatively, use the `--use-system-binaries` option if you '
            'want to use the system wide binaries of ffmpeg/ffprobe/packager.'
        ) from None

    if self._nodes:
      raise RuntimeError('Controller already started!')

    if check_deps:
      # If we are using the hermetic binaries, check the module version.  We
      # must match on the first two digits, but the last one can vary between
      # the two modules.
      if use_hermetic:
        def shorten_version(version: str) -> str:
          """Shorten a version string to the first two digits."""
          components = version.split('.')
          return '.'.join(components[0:2])

        def next_short_version(version: str) -> str:
          """Shorten a version to two digits, then increase the second."""
          components = version.split('.')
          components[1] = str(int(components[1]) + 1)
          return '.'.join(components[0:2])

        streamer_short_version = shorten_version(__version__)
        streamer_binaries_short_version = shorten_version(
            streamer_binaries.__version__)

        if streamer_binaries_short_version != streamer_short_version:
          # This is the recommended install command.  It installs the most
          # recent version of the binary package that matches the current
          # version of streamer itself.  This is much easier to do in nodejs
          # dependencies, because you can use a specifier like "1.2.x", but in
          # Python, you have to use a specifier like ">=1.2,<1.3".
          pip_command = "pip3 install 'shaka-streamer-binaries>={},<{}'".format(
              streamer_short_version, next_short_version(__version__))

          raise VersionError(
              'shaka-streamer-binaries', 'version does not match',
              streamer_short_version,
              exact_match=True,
              addendum='Install with: {}'.format(pip_command))
      else:
        # Check the ffmpeg version.
        _check_command_version('FFmpeg', ['ffmpeg', '-version'],
                               min_versions.FFMPEG)

        # Check the ffprobe version (used for autodetect features).
        _check_command_version('ffprobe', ['ffprobe', '-version'],
                               min_versions.FFMPEG)

        # Check the Shaka Packager version.
        _check_command_version('Shaka Packager', ['packager', '-version'],
                               min_versions.PACKAGER)

    self.hermetic_ffmpeg: Optional[str] = None
    self.hermetic_packager: Optional[str] = None
    if use_hermetic:
      self.hermetic_ffmpeg = streamer_binaries.ffmpeg
      self.hermetic_packager = streamer_binaries.packager
      autodetect.hermetic_ffprobe = streamer_binaries.ffprobe

    # Define resolutions and bitrates before parsing other configs.
    bitrate_config = BitrateConfig(bitrate_config_dict)

    # Now that the definitions have been parsed, register the maps of valid
    # resolutions and channel layouts so that InputConfig and PipelineConfig
    # can be validated accordingly.
    VideoResolution.set_map(bitrate_config.video_resolutions)
    AudioChannelLayout.set_map(bitrate_config.audio_channel_layouts)

    self._input_config = InputConfig(input_config_dict)
    self._pipeline_config = PipelineConfig(pipeline_config_dict)

    if is_http_url(output_location):
      if not self._pipeline_config.segment_per_file:
        raise RuntimeError(
            'For HTTP PUT uploads, the pipeline segment_per_file setting ' +
            'must be set to True!')
    elif is_url(output_location):
      if not ProxyNode.is_understood(output_location):
        url_prefixes = [
            protocol + '://' for protocol in ProxyNode.ALL_SUPPORTED_PROTOCOLS]
        raise RuntimeError(
            'Invalid cloud URL!  Only these are supported: ' +
            ', '.join(url_prefixes))

      if not ProxyNode.is_supported(output_location):
        raise RuntimeError('Missing libraries for cloud URL: ' + output_location)

      if not self._pipeline_config.segment_per_file:
        raise RuntimeError(
            'For cloud uploads, the pipeline segment_per_file setting ' +
            'must be set to True!')

      # These are estimates.  Without constructing the pipeline nodes, we don't
      # know precisely how many output streams there will be.  This is a good
      # guess, though.
      num_video_streams = (
          len(self._pipeline_config.resolutions) *
          len(self._pipeline_config.video_codecs))
      num_audio_streams = (
          len(self._pipeline_config.channel_layouts) *
          len(self._pipeline_config.audio_codecs))
      # Cap the upload pool size at the number of CPUs.
      ideal_pool_size = num_video_streams + num_audio_streams
      max_pool_size = os.cpu_count() or 1  # technically can return None
      pool_size = min(max_pool_size, ideal_pool_size)
      upload_proxy = ProxyNode(output_location, pool_size)
      upload_proxy.start()

      # All the outputs now should be sent to the proxy server instead.
      output_location = upload_proxy.server_location
      self._nodes.append(upload_proxy)
    else:
      # Check if the directory for outputted Packager files exists, and if it
      # does, delete it and remake a new one.
      if os.path.exists(output_location):
        shutil.rmtree(output_location)
      os.mkdir(output_location)

    if self._pipeline_config.low_latency_dash_mode:
      # Check some restrictions on LL-DASH packaging.
      if ManifestFormat.DASH not in self._pipeline_config.manifest_format:
        raise RuntimeError(
            'low_latency_dash_mode is only compatible with DASH ouputs. ' +
            'manifest_format must include DASH')

      if not self._pipeline_config.utc_timings:
        raise RuntimeError(
            'For low_latency_dash_mode, the utc_timings must be set.')

    if self._input_config.inputs:
      # InputConfig contains inputs only.
      self._append_nodes_for_inputs_list(self._input_config.inputs,
                                         output_location)
    else:
      # InputConfig contains multiperiod_inputs_list only.
      if is_url(output_location):
        raise RuntimeError(
            'Direct cloud/HTTP upload is incompatible with multiperiod support.')

      # Create one Transcoder node and one Packager node for each period.
      for i, singleperiod in enumerate(self._input_config.multiperiod_inputs_list):
        sub_dir_name = 'period_' + str(i + 1)
        self._append_nodes_for_inputs_list(singleperiod.inputs,
                                           output_location,
                                           sub_dir_name, i + 1)

      if self._pipeline_config.streaming_mode == StreamingMode.VOD:
        packager_nodes = [node for node in self._nodes if isinstance(node, PackagerNode)]
        self._nodes.append(PeriodConcatNode(
          self._pipeline_config,
          packager_nodes,
          output_location))

    for node in self._nodes:
      node.start()

    return self

  def _append_nodes_for_inputs_list(self, inputs: List[Input],
                                    output_location: str,
                                    period_dir: Optional[str] = None,
                                    index: int = 0) -> None:
    """A common method that creates Transcoder and Packager nodes for a list of Inputs passed to it.

    Args:
      inputs (List[Input]): A list of Input streams.
      output_location (str): A path were the packager will write outputs in.
      period_dir (Optional[str]): A subdirectory name where a single period will be outputted to.
      If passed, this indicates that inputs argument is one period in a list of periods.
      index (int): The index of the current Transcoder/Packager nodes.
    """

    outputs: List[OutputStream] = []
    for input in inputs:
      # External command inputs need to be processed by an additional node
      # before being transcoded.  In this case, the input doesn't have a
      # filename that FFmpeg can read, so we generate an intermediate pipe for
      # that node to write to.  TranscoderNode will then instruct FFmpeg to
      # read from that pipe for this input.
      if input.input_type == InputType.EXTERNAL_COMMAND:
        command_output = Pipe.create_ipc_pipe(self._temp_dir)
        self._nodes.append(ExternalCommandNode(
            input.name, command_output.write_end()))
        # reset the name of the input to be the output pipe path - which the
        # transcoder node will read from - instead of a shell command.
        input.reset_name(command_output.read_end())

      if input.media_type == MediaType.AUDIO:
        for audio_codec in self._pipeline_config.audio_codecs:
          for output_channel_layout in self._pipeline_config.get_channel_layouts():
            # We won't upmix a lower channel count input to a higher one.
            # Skip channel counts greater than the input channel count.
            if input.get_channel_layout() < output_channel_layout:
              continue

            outputs.append(AudioOutputStream(input,
                                             self._temp_dir,
                                             audio_codec,
                                             output_channel_layout))

      elif input.media_type == MediaType.VIDEO:
        for video_codec in self._pipeline_config.video_codecs:
          for output_resolution in self._pipeline_config.get_resolutions():
            # Only going to output lower or equal resolution videos.
            # Upscaling is costly and does not do anything.
            if input.get_resolution() < output_resolution:
              continue

            outputs.append(VideoOutputStream(input,
                                             self._temp_dir,
                                             video_codec,
                                             output_resolution))

      elif input.media_type == MediaType.TEXT:
        if input.name.endswith('.vtt') or input.name.endswith('.ttml'):
          # If the input is a VTT or TTML file, pass it directly to the packager
          # without any intermediate processing or any named pipe.
          # TODO: Test TTML inputs
          skip_transcoding = True  # Bypass transcoder
        else:
          # Otherwise, the input is something like an mkv file with text tracks
          # in it.  These will be extracted by the transcoder and passed in a
          # pipe to the packager.
          skip_transcoding = False

        outputs.append(TextOutputStream(input,
                                        self._temp_dir,
                                        skip_transcoding))

    self._nodes.append(TranscoderNode(inputs,
                                      self._pipeline_config,
                                      outputs,
                                      index,
                                      self.hermetic_ffmpeg))

    # If the inputs list was a period in multiperiod_inputs_list, create a nested directory
    # and put that period in it.
    if period_dir and not is_url(output_location):
      output_location = os.path.join(output_location, period_dir)
      os.mkdir(output_location)

    self._nodes.append(PackagerNode(self._pipeline_config,
                                    output_location,
                                    outputs,
                                    index,
                                    self.hermetic_packager))

  def check_status(self) -> ProcessStatus:
    """Checks the status of all the nodes.

    If one node is errored, this returns Errored; otherwise if one node is running,
    this returns Running; this only returns Finished if all nodes are finished.
    If there are no nodes, this returns Finished.
    """
    if not self._nodes:
      return ProcessStatus.Finished

    value = max(node.check_status().value for node in self._nodes)
    return ProcessStatus(value)

  def stop(self) -> None:
    """Stop all nodes."""
    status = self.check_status()
    for node in self._nodes:
      node.stop(status)
    self._nodes = []

  def is_vod(self) -> bool:
    """Returns True if the pipeline is running in VOD mode.

    :rtype: bool
    """

    return self._pipeline_config.streaming_mode == StreamingMode.VOD

  def is_low_latency_dash_mode(self) -> bool:
    """Returns True if the pipeline is running in LL-DASH mode.

    :rtype: bool
    """

    return self._pipeline_config.low_latency_dash_mode

class VersionError(Exception):
  """A version error for one of Shaka Streamer's external dependencies.

  Raised when a dependency (like FFmpeg) is missing or not new enough to work
  with Shaka Streamer.  See also :doc:`prerequisites`.
  """

  def __init__(self,
               name: str,
               problem: str,
               required_version: str,
               exact_match: bool = False,
               addendum: str = ''):
    or_higher = '' if exact_match else ' or higher'
    message = '{0} {1}! Please install version {2}{3} of {0}.'.format(
        name, problem, required_version, or_higher)
    if addendum:
      message += '\n' + addendum
    super().__init__(message)

def _check_command_version(name: str,
                           command: List[str],
                           minimum_version: Tuple[int, ...]) -> None:
  minimum_version_string = '.'.join(str(x) for x in minimum_version)

  try:
    version_string = str(subprocess.check_output(command))
  except (subprocess.CalledProcessError, OSError) as e:
    if isinstance(e, subprocess.CalledProcessError):
      print(e.stdout, file=sys.stderr)
    raise VersionError(name, 'not found', minimum_version_string) from None

  # Matches two or more numbers (one or more digits each) separated by dots.
  # For example: 4.1.3 or 7.2 or 216.999.8675309
  version_match = re.search(r'[0-9]+(?:\.[0-9]+)+', version_string)

  if version_match:
    version = tuple([int(piece) for piece in version_match.group(0).split('.')])
    if version < minimum_version:
      raise VersionError(name, 'out of date', minimum_version_string)
  else:
    raise RuntimeError(name + ' version could not be parsed!')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import setuptools

import streamer

with open('README.md', 'r') as f:
  long_description = f.read()

setuptools.setup(
  name='shaka-streamer',
  version=streamer.__version__,
  author='Google',
  description='A simple config-file based approach to streaming media.',
  long_description=long_description,
  long_description_content_type='text/markdown',
  url='https://github.com/shaka-project/shaka-streamer',
  packages=setuptools.find_packages(),
  install_requires=[
      'pywin32>=308; platform_system == "Windows"',
      'pyyaml>=6,<7',
      'setproctitle>=1,<2',
      'typing_extensions>=4,<5',
  ],
  scripts=['shaka-streamer'],
  classifiers=[
      'Programming Language :: Python :: 3',
      'License :: OSI Approved :: Apache Software License',
      'Operating System :: POSIX :: Linux',
      'Operating System :: MacOS :: MacOS X',
      'Operating System :: Microsoft :: Windows',
  ],
  # Python 3.9+ tested in GitHub Actions CI
  python_requires='>=3.9',
)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Monkeypatch subprocess on Windows to find .CMD scripts.

Without this patch, subprocess fails to find .CMD scripts on Windows, even
though these are executables and should be treated as such according to the
PATHEXT environment variable.

Many people "work around" this issue on Windows with subprocess's shell=True
argument, but this comes with a risk of shell injection vulnerabilities.

Another solution is to explicitly add ".CMD" to the end of some commands on
Windows, but this breaks portability and requires "if windows" to be scattered
around a codebase.

This monkeypatch allows the caller of subprocess to stop worrying about Windows
nuances and to go back to the security best practice of shell=False.  Any .CMD
script that would be found by the Windows shell will now be found by
subprocess.  And because we're using the standard Windows PATHEXT environment
variable, this can be extended to other types of executable scripts, as well.
"""

import os
import subprocess
import sys


# NOTE: All of the higher-level methods eventually delegate to Popen, so we
# only patch that one method.
#   run => Popen
#   call => Popen
#   check_call => call => Popen
#   check_output => run => Popen

# These environment variables should almost certainly exist, but these are
# defaults in case they are missing.
DEFAULT_PATHEXT = '.COM;.EXE;.BAT;.CMD'
DEFAULT_PATH = r'C:\WINDOWS\system32;C:\WINDOWS'


def resolve(exe):
  """Resolve a command name into a full path to the executable."""

  if '/' in exe or '\\' in exe or ':' in exe:
    # This is a path, so don't modify it.
    return exe

  if '.' in exe:
    # This has an extension already.  Don't search for an extension.
    exe_names = [exe]
  else:
    # This is a command name without an extension, so check for every extension
    # in PATHEXT.
    extensions = os.environ.get('PATHEXT', DEFAULT_PATHEXT).split(';')
    exe_names = [exe + ext for ext in extensions]

  exe_paths = os.environ.get('PATH', DEFAULT_PATH).split(';')

  for path in exe_paths:
    for name in exe_names:
      candidate = os.path.join(path, name)
      if os.access(candidate, os.X_OK):  # If executable
        return candidate

  # Failed to resolve, so return the original name and let Popen fail with a
  # natural-looking error complaining that this command cannot be found.
  return exe


real_Popen = subprocess.Popen


def Popen(args, *more_args, **kwargs):
  """A patch to install over subprocess.Popen."""

  # If the first argument is a list, resolve the command name, which is the
  # first item in the list.
  if isinstance(args, list):
    args[0] = resolve(args[0])

  # Delegate to the real Popen implementation.
  return real_Popen(args, *more_args, **kwargs)


# Only patch win32, but not cygwin.  Cygwin works correctly already.
if sys.platform == 'win32':
  # Patch over Popen.
  subprocess.Popen = Popen # type: ignore
  # Copy the docstring from the real Popen into the patch, so that
  # help(subprocess.Popen) is still relatively sane with this patch installed.
  Popen.__doc__ = real_Popen.__doc__
